@@ -9,7 +9,6 @@ import (
 	"database/sql/driver"
 	"fmt"
 	"io"
-	"strings"
 	"testing"
 	"time"
 )
@@ -57,7 +56,7 @@ func TestCalculateTZ(t *testing.T) {
 		{dbTZ: "+00:00", dbOSTZ: "-07:00", off: -7 * Hour},
 	} {
 		prefix := fmt.Sprintf("%q/%q", tC.dbTZ, tC.dbOSTZ)
-		_, off, err := calculateTZ(tC.dbTZ, tC.dbOSTZ, false)
+		_, off, err := calculateTZ(tC.dbTZ, tC.dbOSTZ, false, nil)
 		t.Logf("tz=%s => off=%d error=%+v", prefix, off, err)
 		if (err == nil) != (tC.err == nil) {
 			t.Errorf("ERR %s: wanted %v, got %v", prefix, tC.err, err)
@@ -68,26 +67,7 @@ func TestCalculateTZ(t *testing.T) {
 }
 
 func TestTZName(t *testing.T) {
-	f := func(dbTZ string) (*time.Location, error) {
-		t.Log("try", dbTZ)
-		tz, err := time.LoadLocation(dbTZ)
-		if err != nil {
-			for _, nm := range tzNames {
-				if strings.EqualFold(nm, dbTZ) {
-					tz, err = time.LoadLocation(nm)
-					break
-				}
-			}
-		}
-		if err == nil {
-			if tz == nil {
-				tz = time.UTC
-			}
-			return tz, nil
-		}
-		return nil, err
-	}
-
+	t.Log("tzNamesLC", tzNamesLC)
 	chicago, err := time.LoadLocation("America/Chicago")
 	if err != nil {
 		t.Fatal(err)
@@ -110,7 +90,7 @@ func TestTZName(t *testing.T) {
 		"AMERICA/CHICAGO":  {chicago, nil},
 		"AMERICA/NeW_yORk": {newYork, nil},
 	} {
-		got, err := f(in)
+		got, err := findProperTZName(in)
 		if err != want.error {
 			t.Errorf("%q got error %+v, wanted %v", in, err, want.error)
 		} else if got != want.Location && got.String() != want.Location.String() {
