@@ -13,12 +13,15 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
+	"io"
+	"log/slog"
 	"runtime"
 	"strings"
 	"sync"
 	"testing"
 	"time"
 
+	"github.com/UNO-SOFT/zlog/v2"
 	godror "github.com/godror/godror"
 	"github.com/google/go-cmp/cmp"
 	"golang.org/x/sync/errgroup"
@@ -32,8 +35,9 @@ var _ godror.ObjectScanner = new(MyTable)
 // MYRecord represents TEST_PKG_TYPES.MY_RECORD
 type MyRecord struct {
 	*godror.Object
-	Txt string
-	ID  int64
+	Txt, AClob string
+	ID         int64
+	DT         time.Time
 }
 
 type coder interface{ Code() int }
@@ -55,6 +59,23 @@ func (r *MyRecord) Scan(src interface{}) error {
 	}
 	r.Txt = txt.(string)
 
+	dt, err := obj.Get("DT")
+	if err != nil {
+		return err
+	}
+	r.DT = dt.(time.Time)
+
+	aclob, err := obj.Get("ACLOB")
+	if err != nil {
+		return err
+	}
+	if aclob != nil {
+		if lob, ok := aclob.(*godror.Lob); ok && lob != nil {
+			b, err := io.ReadAll(lob)
+			r.AClob = string(b)
+			return err
+		}
+	}
 	return nil
 }
 
@@ -85,6 +106,24 @@ func (r MyRecord) WriteObject() error {
 		r.SetAttribute("TXT", &data)
 	}
 
+	if !r.DT.IsZero() {
+		err = r.GetAttribute(&data, "DT")
+		if err != nil {
+			return err
+		}
+		data.SetTime(r.DT)
+		r.SetAttribute("DT", &data)
+	}
+
+	if r.AClob != "" {
+		err = r.GetAttribute(&data, "ACLOB")
+		if err != nil {
+			return err
+		}
+		data.SetBytes([]byte(r.AClob))
+		r.SetAttribute("ACLOB", &data)
+	}
+
 	return nil
 }
 
@@ -173,7 +212,11 @@ func createPackages(ctx context.Context) error {
 	TYPE my_record IS RECORD (
 		id    NUMBER(5),
 		other test_pkg_types.my_other_record,
-		txt   VARCHAR2(200)
+		txt   VARCHAR2(200),
+		dt    DATE,
+		aclob CLOB,
+		child test_pkg_types.my_other_record,
+		int32 PLS_INTEGER
 	);
 	TYPE my_table IS TABLE OF my_record;
 
@@ -225,6 +268,9 @@ func createPackages(ctx context.Context) error {
 	BEGIN
 		rec.id := id;
 		rec.txt := txt;
+		rec.dt := SYSDATE;
+		--DBMS_LOB.createtemporary(rec.aclob, TRUE);
+		--DBMS_LOB.write(rec.aclob, LENGTH(txt)*2+1, 1, txt||'+'||txt);
 	END test_record;
 
 	PROCEDURE test_record_in (
@@ -232,7 +278,11 @@ func createPackages(ctx context.Context) error {
 	) IS
 	BEGIN
 		rec.id := rec.id + 1;
-		rec.txt := rec.txt || ' changed';
+		rec.txt := rec.txt || ' changed '||TO_CHAR(rec.dt, 'YYYY-MM-DD HH24:MI:SS');
+		rec.dt := SYSDATE;
+		--IF rec.aclob IS NOT NULL AND DBMS_LOB.isopen(rec.aclob) <> 0 THEN
+		--  DBMS_LOB.writeappend(rec.aclob, 7, ' changed');
+		--END IF;
 	END test_record_in;
 
 	FUNCTION test_table (
@@ -251,6 +301,7 @@ func createPackages(ctx context.Context) error {
 				level <= x
 		) LOOP
 			item.id := c.lev;
+			item.dt := SYSDATE;
 			item.txt := 'test - ' || ( c.lev * 2 );
 			tb.extend();
 			tb(tb.count) := item;
@@ -280,6 +331,9 @@ func createPackages(ctx context.Context) error {
 		FOR i IN 1..3 LOOP
 			res_list.extend();
 			rec.id := i;
+			rec.rec.id := 2*i+1;  --initialize sub-record for #283
+			rec.rec.dt := SYSDATE;
+			--test_record(rec.id, rec.id, rec.rec);
 			res_list(res_list.count) := rec;
 		END LOOP;
 
@@ -295,6 +349,14 @@ func createPackages(ctx context.Context) error {
 		}
 	}
 
+	errs, err := godror.GetCompileErrors(ctx, testDb, false)
+	if err != nil {
+		return err
+	}
+	for _, err := range errs {
+		return err
+	}
+
 	return nil
 }
 
@@ -305,10 +367,12 @@ func dropPackages(ctx context.Context) {
 
 type objectStruct struct {
 	godror.ObjectTypeName `godror:"test_pkg_types.my_record" json:"-"`
-	ID                    int32  `godror:"ID"`
-	Txt                   string `godror:"TXT"`
+	ID                    int32         `godror:"ID"`
+	Txt                   string        `godror:"TXT"`
+	DT                    time.Time     `godror:"DT"`
+	AClob                 string        `godror:"ACLOB"`
+	NInt32                sql.NullInt32 `godror:"INT32"`
 }
-
 type sliceStruct struct {
 	godror.ObjectTypeName `json:"-"`
 	ObjSlice              []objectStruct `godror:",type=test_pkg_types.my_table"`
@@ -375,6 +439,10 @@ func TestPlSqlTypes(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
+	if false && Verbose {
+		defer tl.enableLogging(t)()
+		godror.SetLogger(zlog.NewT(t).SLog())
+	}
 
 	t.Run("Struct", func(t *testing.T) {
 		var s objectStruct
@@ -389,7 +457,10 @@ func TestPlSqlTypes(t *testing.T) {
 	t.Run("Slice", func(t *testing.T) {
 		s := sliceStruct{ObjSlice: []objectStruct{{ID: 1, Txt: "first"}}}
 		const qry = `begin test_pkg_sample.test_table_in(:1); end;`
-		_, err := cx.ExecContext(ctx, qry, sql.Out{Dest: &s, In: true})
+		_, err := cx.ExecContext(ctx,
+			qry,
+			sql.Out{Dest: &s, In: true},
+		)
 		if err != nil {
 			t.Fatalf("%s: %+v", qry, err)
 		}
@@ -453,8 +524,8 @@ func TestPlSqlTypes(t *testing.T) {
 				t.Fatal(err)
 			}
 
-			if rec != tCase.want {
-				t.Errorf("%s: record got %v, wanted %v", tName, rec, tCase.want)
+			if d := cmp.Diff(rec.String(), tCase.want.String()); d != "" {
+				t.Errorf("%s: record got %v, wanted %v\n%s", tName, rec, tCase.want, d)
 			}
 		}
 	})
@@ -471,9 +542,9 @@ func TestPlSqlTypes(t *testing.T) {
 			in      MyRecord
 			wantID  int64
 		}{
-			"zeroValues": {in: MyRecord{}, wantID: 1, wantTxt: " changed"},
-			"default":    {in: MyRecord{ID: 1, Txt: "test"}, wantID: 2, wantTxt: "test changed"},
-			"emptyTxt":   {in: MyRecord{ID: 2, Txt: ""}, wantID: 3, wantTxt: " changed"},
+			"zeroValues": {in: MyRecord{}, wantID: 1, wantTxt: " changed "},
+			"default":    {in: MyRecord{ID: 1, Txt: "test"}, wantID: 2, wantTxt: "test changed "},
+			"emptyTxt":   {in: MyRecord{ID: 2, Txt: ""}, wantID: 3, wantTxt: " changed "},
 		} {
 
 			obj, err := objType.NewObject()
@@ -1186,7 +1257,7 @@ func TestObjectTypeClose(t *testing.T) {
 		return nil
 	}
 
-	const maxConn = maxSessions * 2
+	maxConn := maxSessions * 2
 	for j := 0; j < 5; j++ {
 		t.Logf("Run %d group\n", j)
 		var start sync.WaitGroup
@@ -1251,7 +1322,7 @@ func TestSubObjectTypeClose(t *testing.T) {
 		return nil
 	}
 
-	const maxConn = maxSessions * 2
+	maxConn := maxSessions * 2
 	for j := 0; j < 5; j++ {
 		t.Logf("Run %d group\n", j)
 		var start sync.WaitGroup
@@ -1394,7 +1465,7 @@ END;`},
 			nil,
 		}
 		if d := cmp.Diff(m, want); d != "" {
-			t.Errorf(d)
+			t.Error(d)
 		}
 	})
 
@@ -1548,7 +1619,7 @@ END;`},
 			},
 		}
 		if d := cmp.Diff(m, want); d != "" {
-			t.Errorf(d)
+			t.Error(d)
 		}
 	})
 
@@ -1760,6 +1831,104 @@ END;`,
 	}
 }
 
+func TestInputWithNativeSlice(t *testing.T) {
+	ctx, cancel := context.WithTimeout(testContext("InputWithNativeSlice"), 10*time.Second)
+	defer cancel()
+	t.Logf("dbstats: %#v", testDb.Stats())
+	name := "test_ons" + tblSuffix
+
+	cleanup := func() {
+		testDb.Exec("DROP PROCEDURE " + name + "_proc")
+		testDb.Exec("DROP TYPE " + name + "_list")
+	}
+	cleanup()
+	defer cleanup()
+	create_query := []string{
+		`CREATE OR REPLACE TYPE ` + name + `_list AS TABLE OF NUMBER;`,
+		`CREATE OR REPLACE PROCEDURE ` + name + `_proc(p_in IN ` + name + `_list, p_out OUT NUMBER) IS
+         v_result NUMBER := 0;
+         BEGIN
+           FOR v IN 1..p_in.COUNT LOOP
+             v_result := v_result + p_in(v);
+           END LOOP;
+		
+           p_out := v_result;
+         END;`,
+	}
+
+	for _, q := range create_query {
+		if _, err := testDb.ExecContext(ctx, q); err != nil {
+			t.Fatalf("%s: %+v", q, err)
+		}
+	}
+	if ces, err := godror.GetCompileErrors(ctx, testDb, false); err != nil {
+		t.Fatal(err)
+	} else if len(ces) != 0 {
+		t.Fatal(ces)
+	}
+
+	conn, err := testDb.Conn(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer conn.Close()
+	testCon, err := godror.DriverConn(ctx, conn)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer testCon.Close()
+	if err = testCon.Ping(ctx); err != nil {
+		t.Fatal(err)
+	}
+
+	qry := "CALL " + name + "_proc(:1, :2)"
+	stmt, err := conn.PrepareContext(ctx, qry)
+	if err != nil {
+		t.Fatalf("%s: %+v", qry, err)
+	}
+	defer stmt.Close()
+
+	cCt, err := testCon.GetObjectType(strings.ToUpper(name + "_list"))
+	if err != nil {
+		t.Fatal(err)
+	}
+	t.Log("ObjectType:", cCt)
+
+	cC, err := cCt.NewCollection()
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer cC.Close()
+	fSlice := []float64{1, 3.14}
+	var want float64
+	iSlice := make([]interface{}, len(fSlice))
+	for i, f := range fSlice {
+		iSlice[i] = f
+		want += f
+	}
+	if err = cC.FromSlice(iSlice); err != nil {
+		t.Fatal(err)
+	}
+	length, err := cC.Len()
+	if err != nil {
+		t.Fatal(err)
+	}
+	t.Log("length of cC:", length)
+
+	var sum float64
+	if _, err = stmt.ExecContext(ctx,
+		cC,
+		sql.Out{Dest: &sum},
+	); err != nil {
+		t.Errorf("%s: %+v", qry, err)
+	}
+
+	t.Log("Sum:", sum)
+	if fmt.Sprintf("%f", sum) != fmt.Sprintf("%f", want) {
+		t.Errorf("got %f, wanted %f", sum, want)
+	}
+}
+
 func TestPlSQLNumSlice(t *testing.T) {
 	ctx, cancel := context.WithTimeout(testContext("PlSQLNumSlice"), 30*time.Second)
 	defer cancel()
@@ -1859,3 +2028,665 @@ END;`,
 	}
 	t.Log(s, len(s))
 }
+
+func TestXMLType(t *testing.T) {
+	ctx, cancel := context.WithTimeout(testContext("XMLType"), 30*time.Second)
+	defer cancel()
+
+	const qry = "SELECT 'text' AS text, XMLElement(\"Date\", SYSDATE) AS xml FROM DUAL"
+	var txt, xml string
+	if err := testDb.QueryRowContext(ctx, qry).Scan(&txt, &xml); err != nil {
+		t.Fatalf("%s: %+v", qry, err)
+	}
+	t.Logf("txt=%q xml=%q", txt, xml)
+
+	rows, err := testDb.QueryContext(ctx, qry)
+	if err != nil {
+		t.Fatal(err)
+	}
+	types, err := rows.ColumnTypes()
+	if err != nil {
+		t.Fatal(err)
+	}
+	t.Logf("coltype: %v, %v", types[0], types[1])
+	defer rows.Close()
+	for rows.Next() {
+		var i, j interface{}
+		if err = rows.Scan(&i, &j); err != nil {
+			t.Fatal(err)
+		}
+		t.Logf("column[0]: (%[1]T) %#[1]v", i)
+		t.Logf("column[1]: (%[1]T) %#[1]v", j)
+
+	}
+}
+
+func TestBigXMLType(t *testing.T) {
+	ctx, cancel := context.WithTimeout(testContext("BigXMLType"), 30*time.Second)
+	defer cancel()
+	{
+		_, _ = testDb.ExecContext(context.Background(), "DROP TABLE test_xml")
+		const qry = "create table test_xml(id number,data xmltype)"
+		if _, err := testDb.ExecContext(ctx, qry); err != nil {
+			t.Fatalf("%q: %+v", qry, err)
+		}
+		defer testDb.ExecContext(context.Background(), "DROP TABLE test_xml")
+	}
+	// defer db.Exec("drop table test_xml")
+
+	const want = `<root>
+  <person
+  firstname="Marsiella"
+  lastname="Schonfeld"
+  city="Arbil"
+  country="Nicaragua"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person2
+  firstname="Emilia"
+  lastname="Glovsky"
+  city="Cali"
+  country="Cyprus"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person3
+  firstname="Dari"
+  lastname="Mike"
+  city="Mata-Utu"
+  country="Belize"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person4
+  firstname="Ericka"
+  lastname="Swanhildas"
+  city="Wichita"
+  country="Somalia"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person5
+  firstname="Luci"
+  lastname="Gualtiero"
+  city="Iquique"
+  country="Korea, Republic of"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person6
+  firstname="Ofilia"
+  lastname="Ailyn"
+  city="Sydney"
+  country="Tanzania, United Republic of"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person7
+  firstname="Edith"
+  lastname="Dannye"
+  city="Charlotte Amalie"
+  country="Swaziland"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person8
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="Portland"
+  country="Oman"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person9
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="Portland"
+  country="Oman"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person10
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="Portland"
+  country="Oman"
+  firstname2="Ardys"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person11
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="Portland"
+  country="Oman"
+  firstname2="Ardys"
+  lastname2="中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person12
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="Portland"
+  country="Oman"
+  firstname2="Ardys"
+  lastname2="中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person13
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="Portland"
+  country="Oman"
+  firstname2="中文测试"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person14
+  firstname="Gratia"
+  lastname="Ephrem"
+  city="中文测试中文测试中文测试中文测试"
+  country="中文测试中文测试中文测试中文测试"
+  firstname2="中文测试中文测试中文测试中文测试"
+  lastname2="中文测试中文测试中文测试中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person15
+  firstname="中文测试中文测试中文测试中文测试中文测试中文测试"
+  lastname="中文测试中文测试中文测试中文测试"
+  city="中文测试中文测试中文测试中文测试"
+  country="中文测试中文测试中文测试中文测试"
+  firstname2="中文测试中文测试中文测试中文测试"
+  lastname2="中文测试中文测试中文测试中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person16
+  firstname="中文测试中文测试中文测试中文测试"
+  lastname="中文测试中文测试"
+  city="中文测试中文测试中文测试中文测试"
+  country="中文测试中文测试中文测试中文测试"
+  firstname2="中文测试中文测试"
+  lastname2="中文测试中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person17
+  firstname="中文测试"
+  lastname="中文测试"
+  city="中文测试"
+  country="中文测试"
+  firstname2="中文测试"
+  lastname2="中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person18
+  firstname="中文测试中文测试"
+  lastname="中文测试中文测试"
+  city="中文测试中文测试"
+  country="中文测试中文测试"
+  firstname2="中文测试中文测试"
+  lastname2="中文测试中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person19
+  firstname="中文测试中文测试"
+  lastname="中文测试中文测试"
+  city="中文测试中文测试"
+  country="中文测试中文测试"
+  firstname2="中文测试中文测试"
+  lastname2="Doig"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person20
+  firstname="中文测试中文测试"
+  lastname="中文测试中文测试"
+  city="中文测试中文测试"
+  country="中文测试中文测试"
+  firstname2="中文测试中文测试"
+  lastname2="中文测试中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <person21
+  firstname="Gratia"
+  lastname="中文测试中文测试"
+  city="中文测试中文测试"
+  country="中文测试中文测试"
+  firstname2="中文测试中文测试"
+  lastname2="中文测试中文测试"
+  email="Ardys.Doig@yopmail.com"
+  />
+  <random>38</random>
+  <random_float>28.038</random_float>
+  <bool>true</bool>
+  <date>1983-01-07</date>
+  <regEx>hell0</regEx>
+  <enum>online</enum>
+  <elt>Cherilyn</elt><elt>Frieda</elt><elt>Kimberley</elt><elt>Celestyna</elt><elt>Ethel</elt>  
+  <Ardys>
+    <age>34</age>
+  </Ardys>
+</root>`
+	{
+		const qry = "insert into test_xml(id,data) values (:1,:2)"
+		if _, err := testDb.ExecContext(ctx, qry, 1, want); err != nil {
+			t.Fatalf("insert %s", err)
+		}
+	}
+	const qry = "select A.id, A.data.getClobVal() AS data FROM test_xml A"
+	rows, err := testDb.QueryContext(ctx, qry)
+	if err != nil {
+		t.Fatalf("%s: %+v", qry, err)
+	}
+	defer rows.Close()
+	cols, err := rows.ColumnTypes()
+	if err != nil {
+		t.Fatal(err)
+	}
+	for _, c := range cols {
+		t.Logf("column %q: %q", c.Name(), c.DatabaseTypeName())
+	}
+	sRepl := strings.NewReplacer(" ", "", "\n", "", "\t", "")
+	for rows.Next() {
+		var id, data string
+		if err := rows.Scan(&id, &data); err != nil {
+			t.Fatal("scan: %+w", err)
+		}
+		t.Logf("id=%q data=%d", id, len(data))
+		if d := cmp.Diff(sRepl.Replace(data), sRepl.Replace(want)); d != "" {
+			t.Error(d)
+		}
+	}
+	if err := rows.Err(); err != nil {
+		t.Fatal(err)
+	}
+
+}
+
+func TestArrayOfRecords(t *testing.T) {
+	ctx, cancel := context.WithTimeout(testContext("ArrayOfRecords"), 10*time.Second)
+	defer cancel()
+	name := "test_aor"
+
+	cleanup := func() {
+		testDb.Exec("DROP PACKAGE " + name + "_pkg")
+	}
+	cleanup()
+	defer cleanup()
+	create_query := []string{
+		`CREATE OR REPLACE PACKAGE ` + name + `_pkg AS 
+  TYPE ` + name + `_rt IS RECORD (
+    ratecode_id PLS_INTEGER, 
+    ratecode VARCHAR2(20), 
+    ratecode_name VARCHAR2(100), 
+    rate_structure VARCHAR2(20)
+  );
+  TYPE ` + name + `_tt IS TABLE OF ` + name + `_rt;
+
+  PROCEDURE ` + name + `_proc(p_in IN ` + name + `_tt, p_out OUT VARCHAR2);
+END;`,
+
+		`CREATE OR REPLACE PACKAGE BODY ` + name + `_pkg AS
+  PROCEDURE ` + name + `_proc(p_in IN ` + name + `_tt, p_out OUT VARCHAR2) IS
+    v_result VARCHAR2(1000);
+    v_idx PLS_INTEGER;
+  BEGIN
+    v_idx := p_in.FIRST;
+    WHILE v_idx IS NOT NULL LOOP
+      v_result := v_result || p_in(v_idx).ratecode_id||'='||p_in(v_idx).ratecode||', ';
+      v_idx := p_in.NEXT(v_idx);
+    END LOOP;
+    p_out := v_result;
+  END;
+END;`,
+	}
+
+	for _, q := range create_query {
+		if _, err := testDb.ExecContext(ctx, q); err != nil {
+			t.Fatalf("%s: %+v", q, err)
+		}
+		if ces, err := godror.GetCompileErrors(ctx, testDb, false); err != nil {
+			t.Fatal(err)
+		} else if len(ces) != 0 {
+			t.Log(q)
+			t.Fatal(ces)
+		}
+	}
+
+	conn, err := testDb.Conn(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer conn.Close()
+
+	qry := `begin ` + name + `_pkg.` + name + `_proc(:1, :2); end;`
+	var resp string
+	if _, err := conn.ExecContext(ctx, qry, RatePlan{RateCodes: []RateCode{
+		{RateCodeId: 1, RateCode: "one"},
+		{RateCodeId: 2, RateCode: "two"},
+	}},
+		sql.Out{Dest: &resp},
+	); err != nil {
+		t.Fatalf("%s: %+v", qry, err)
+	}
+	t.Log(resp)
+}
+
+type RatePlan struct {
+	godror.ObjectTypeName `godror:"test_aor_pkg.test_aor_tt" json:"-"`
+	RateCodes             []RateCode `json:"rateCodes"`
+}
+
+type RateCode struct {
+	godror.ObjectTypeName `godror:"test_aor_pkg.test_aor_rt" json:"-"`
+	RateCodeId            int    `godror:"RATECODE_ID" json:"rateCodeId" db:"RATECODE_ID"`
+	RateCode              string `json:"rateCode" db:"RATECODE"`
+	RateCodeName          string `godror:"RATECODE_NAME" json:"rateCodeName" db:"RATECODE_NAME"`
+	RateStucture          string `godror:"RATE_STRUCTURE" json:"rateStucture" db:"RATE_STRUCTURE"`
+}
+
+type parentObject struct {
+	godror.ObjectTypeName `godror:"test_parent_ot"`
+	ID                    int         `godror:"ID"`
+	Child                 childObject `godror:"CHILD"`
+}
+type childObject struct {
+	godror.ObjectTypeName `godror:"test_child_ot"`
+	ID                    int    `godror:"ID"`
+	Name                  string `godror:"NAME"`
+}
+
+func TestIssue319(t *testing.T) {
+	ctx, cancel := context.WithTimeout(testContext("Issue319"), 10*time.Second)
+	defer cancel()
+	if Verbose {
+		godror.SetLogger(zlog.NewT(t).SLog())
+		defer godror.SetLogger(slog.Default())
+	}
+
+	cleanup := func() {
+		for _, qry := range []string{
+			"DROP TYPE test_parent_ot",
+			"DROP TYPE test_child_ot",
+		} {
+			if _, err := testDb.ExecContext(context.Background(), qry); err != nil {
+				t.Logf("%s: %+v", qry, err)
+			}
+		}
+	}
+
+	cleanup()
+	defer cleanup()
+	for _, qry := range []string{
+		"CREATE OR REPLACE TYPE test_child_ot AS OBJECT (id NUMBER(3), name VARCHAR2(128));",
+		"CREATE OR REPLACE TYPE test_parent_ot AS OBJECT (id NUMBER(3), child test_child_ot);",
+	} {
+		if _, err := testDb.ExecContext(ctx, qry); err != nil {
+			t.Fatalf("%s: %+v", qry, err)
+		}
+	}
+	P, err := godror.ParseDSN(testConStr)
+	if err != nil {
+		t.Fatal(err)
+	}
+	P.MinSessions, P.MaxSessions = 1, 1
+	t.Logf("params: %s", P)
+	db := sql.OpenDB(godror.NewConnector(P))
+	defer db.Close()
+
+	Exec := func(ctx context.Context, qry string, params ...any) error {
+		tx, err := db.BeginTx(ctx, nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		defer tx.Rollback()
+		stmt, err := tx.PrepareContext(ctx, qry)
+		if err != nil {
+			t.Fatalf("prepare %s: %+v", qry, err)
+		}
+		defer stmt.Close()
+		_, err = stmt.ExecContext(ctx, params...)
+		return err
+	}
+
+	for i := 0; i < 5; i++ {
+		c := childObject{ID: 1, Name: "test"}
+		p := parentObject{ID: 1, Child: c}
+		var res int
+
+		pre := db.Stats()
+		if err := Exec(ctx, `DECLARE
+  v_parent test_parent_ot;
+BEGIN
+  v_parent := :p;
+  v_parent.id := v_parent.id + 1;
+  :res := v_parent.id;
+END;`,
+			sql.Named("p", p),
+			sql.Named("res", sql.Out{Dest: &res}),
+		); err != nil {
+			t.Fatalf("%d: %+v", i, err)
+		}
+		t.Logf("%d. result: %d", i, res)
+		post := db.Stats()
+		t.Logf("\npre: %+v\npost: %+v", pre, post)
+		if post.OpenConnections > pre.OpenConnections+1 {
+			t.Errorf("session leakage: was %d, have %d open connections", pre.OpenConnections, post.OpenConnections)
+		}
+	}
+}
+
+func TestObjLobClose(t *testing.T) {
+	const dropQry = `DROP TYPE test_clob_ot`
+	cleanup := func() { testDb.ExecContext(context.Background(), dropQry) }
+	cleanup()
+	const qry = `create or replace type test_clob_ot as object(
+  id number,
+  value clob
+)`
+	ctx, cancel := context.WithTimeout(testContext("ObjLobClose"), 10*time.Second)
+	defer cancel()
+	if _, err := testDb.ExecContext(ctx, qry); err != nil {
+		t.Fatalf("%s: %+v", qry, err)
+	}
+	defer cleanup()
+
+	Exec := func(ctx context.Context, db *sql.DB, query string, variables ...any) error {
+		tx, err := db.BeginTx(ctx, nil)
+		if err != nil {
+			return err
+		}
+		defer tx.Rollback()
+
+		stmt, err := tx.PrepareContext(ctx, query)
+		if err != nil {
+			return err
+		}
+		defer stmt.Close()
+
+		_, err = stmt.ExecContext(ctx, variables...)
+		return err
+	}
+
+	type clob struct {
+		godror.ObjectTypeName `godror:"test_clob_ot"`
+		Id                    float64 `godror:"ID"`
+		Value                 string  `godror:"VALUE"`
+	}
+
+	testConnectInClob := func(ctx context.Context, db *sql.DB, i float64) {
+		t.Log("start In clob", i)
+		var c clob
+		c.Id = 1
+		c.Value = `test`
+
+		var res float64
+		err := Exec(ctx, db, `
+	declare
+		v_child test_clob_ot;
+	begin
+		v_child := :c;
+		v_child.id := v_child.id + 5;
+		:res := v_child.id;
+	end;`,
+			sql.Named(`c`, c),
+			sql.Named(`res`, sql.Out{In: false, Dest: &res}),
+		)
+		t.Log("result", i, res, err)
+	}
+
+	testConnectOutClob := func(ctx context.Context, db *sql.DB, i float64) {
+		t.Log("start Out clob", i)
+		var c clob
+
+		err := Exec(ctx, db, `
+	declare
+		v_child test_clob_ot := test_clob_ot(id => null, value => null);
+	begin
+		v_child.id :=:i + 5;
+		v_child.value := 'hello world ' || v_child.id;
+		:v_res := v_child;
+	end;`,
+			sql.Named(`i`, i),
+			sql.Named(`v_res`, sql.Out{In: false, Dest: &c}),
+		)
+		t.Log("result", i, c, err)
+	}
+
+	P, err := godror.ParseDSN(testConStr)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	P.PoolParams.SessionTimeout = 30 * time.Second
+	P.PoolParams.WaitTimeout = 10 * time.Second
+	P.PoolParams.MaxLifeTime = 30 * time.Second
+	P.PoolParams.SessionIncrement = 1
+	P.PoolParams.MinSessions = 1
+	P.PoolParams.MaxSessions = 3
+	db := sql.OpenDB(godror.NewConnector(P))
+	defer db.Close()
+
+	for i := 0; i < 10; i++ {
+		testConnectInClob(ctx, db, float64(i))
+	}
+
+	for i := 0; i < 10; i++ {
+		testConnectOutClob(ctx, db, float64(i))
+	}
+}
+
+type I323Child struct {
+	godror.ObjectTypeName `godror:"I323CHILD"`
+	ID                    float64 `godror:"ID" json:"id"`
+	Name                  string  `godror:"NAME" json:"name"`
+}
+
+type I323ChildArray struct {
+	godror.ObjectTypeName
+	ChildArray []I323Child `godror:",type=I323CHILDARRAY" json:"childArray"`
+}
+
+type I323Parent struct {
+	godror.ObjectTypeName `godror:"I323PARENT"`
+	Id                    float64        `godror:"ID" json:"id"`
+	Name                  string         `godror:"NAME" json:"name"`
+	ChildArray            I323ChildArray `godror:",type=I323CHILDARRAY"`
+}
+
+type I323ParentArray struct {
+	godror.ObjectTypeName
+	ParentArray []I323Parent `godror:",type=I323PARENTARRAY" json:"parentArray"`
+}
+
+type I323Grand struct {
+	godror.ObjectTypeName `godror:"I323GRAND"`
+	ParentArray           I323ParentArray `godror:",type=I323PARENTARRAY"`
+}
+
+func TestIssue323(t *testing.T) {
+	drop := func() {
+		for _, qry := range []string{
+			"DROP TYPE I323GRAND FORCE",
+			"DROP TYPE I323PARENTARRAY FORCE",
+			"DROP TYPE I323PARENT FORCE",
+			"DROP TYPE I323CHILDARRAY FORCE",
+			"DROP TYPE I323CHILD FORCE",
+		} {
+			if _, err := testDb.ExecContext(context.Background(), qry); err != nil {
+				t.Logf("%s: %+v", qry, err)
+			}
+		}
+	}
+	drop()
+	defer drop()
+
+	ctx, cancel := context.WithTimeout(testContext("Issue323"), 10*time.Second)
+	defer cancel()
+	for _, qry := range []string{
+		"CREATE OR REPLACE TYPE I323CHILD AS OBJECT (ID NUMBER, NAME VARCHAR2(100))",
+		"CREATE OR REPLACE TYPE I323CHILDARRAY AS TABLE OF I323CHILD",
+		"CREATE OR REPLACE TYPE I323PARENT AS OBJECT(ID NUMBER, NAME varchar2(100), childArray I323CHILDARRAY)",
+		"CREATE OR REPLACE TYPE I323PARENTARRAY AS TABLE OF I323PARENT",
+		"CREATE OR REPLACE TYPE I323Grand AS OBJECT(ParentArray I323ParentArray)",
+	} {
+		if _, err := testDb.ExecContext(ctx, qry); err != nil {
+			t.Fatalf("%s: %+v", qry, err)
+		}
+	}
+
+	qry := `
+	declare
+		procedure test(
+			p_grand out I323Grand
+		) is
+			v_childArray I323ChildArray := I323ChildArray();
+		begin
+			p_grand :=  I323Grand(null);
+			p_grand.parentArray := I323ParentArray();
+			for i in 1..2 loop
+				p_grand.parentArray.extend;
+				p_grand.parentArray(p_grand.parentArray.count) := I323Parent(i, 'value ' || i, null);
+				p_grand.ParentArray(p_grand.parentArray.count).childArray := I323ChildArray();
+				v_childArray := I323ChildArray();
+				for j in 1..2 loop
+					v_childArray.extend;
+					v_childArray(v_childArray.count) := I323Child(i*100 + j, 'coord '||i||'/'||j);
+				end loop;
+				p_grand.ParentArray(p_grand.parentArray.count).ChildArray := v_childArray;
+			end loop;
+		end;
+	begin
+		test(
+			p_grand => :g
+		);
+	end;`
+
+	if Verbose {
+		godror.SetLogger(zlog.NewT(t).SLog())
+		defer godror.SetLogger(slog.Default())
+	}
+
+	var grand I323Grand
+	if _, err := testDb.ExecContext(ctx, qry,
+		sql.Named(`g`, sql.Out{In: false, Dest: &grand}),
+	); err != nil {
+		t.Fatalf("%s: %+v", qry, err)
+	}
+	p, _ := json.Marshal(grand)
+	t.Log("grand:", string(p))
+
+	pa := grand.ParentArray.ParentArray
+	if pa[0].ChildArray.ChildArray[0].ID ==
+		pa[1].ChildArray.ChildArray[0].ID {
+		t.Errorf("pa[0]=%#v\n==\n%#v=pa[1]", pa[0], pa[1])
+	}
+}
